/*
 * Gimbal
 * Copyright (C) 2024 moonleay
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package net.moonleay.gimbal.client.editor

import net.minecraft.client.MinecraftClient
import net.minecraft.text.Text
import net.moonleay.gimbal.client.util.ChatUtil
import net.moonleay.gimbal.editor.ServerEditorManager
import net.moonleay.gimbal.editor.state.EditorState
import net.moonleay.gimbal.editor.state.GimbalPolicyType
import net.moonleay.gimbal.editor.state.GimbalServerState
import net.moonleay.gimbal.editor.state.mode.Capability
import net.moonleay.gimbal.editor.state.mode.Mode
import net.moonleay.gimbal.editor.state.mode.ModeModifier
import net.moonleay.gimbal.editor.util.EditorUtil
import net.moonleay.gimbal.networking.GimbalClient

object ClientEditor {
    private var POLICY = GimbalPolicyType.NOT_PRESENT

    private var CURRENT_MODE = Mode.NORMAL
    private var TEMP_DISABLED_MODE = Mode.UNKNOWN

    private val CURRENT_MODE_MODIFIER = mutableListOf<ModeModifier>()
    private val TEMP_DISABLED_MODIFIERS = mutableListOf<ModeModifier>()

    private val DISABLED_MODIFIERS_STORAGE = mutableListOf<ModeModifier>()


    fun onConnectedToNewWorld() {
        POLICY = GimbalPolicyType.NOT_PRESENT
        if (TEMP_DISABLED_MODE == Mode.UNKNOWN) {
            TEMP_DISABLED_MODE =
                CURRENT_MODE
        }
        CURRENT_MODE = Mode.INSERT
        DISABLED_MODIFIERS_STORAGE.addAll(
            CURRENT_MODE_MODIFIER
        )
        CURRENT_MODE_MODIFIER.clear()
    }

    fun onPolicyReceived(data: GimbalServerState, shouldUpdateServer: Boolean = true) {
        POLICY = data.policyType
        if (data.policyType == GimbalPolicyType.ALLOWED) {
            if (TEMP_DISABLED_MODE != Mode.UNKNOWN) {
                CURRENT_MODE =
                    TEMP_DISABLED_MODE
                TEMP_DISABLED_MODE = Mode.UNKNOWN
            }
            CURRENT_MODE_MODIFIER.addAll(
                DISABLED_MODIFIERS_STORAGE
            )
            DISABLED_MODIFIERS_STORAGE.clear()
            onUpdated(shouldUpdateServer)

            if (!shouldUpdateServer) {
                val player = MinecraftClient.getInstance().player
                if (player == null) {
                    println("Player is null!")
                    return
                }
                ServerEditorManager.updateEditorState(
                    MinecraftClient.getInstance().player!!.uuid,
                    this.getClientState()
                )
            }
        }
    }

    fun isAllowed(): Boolean {
        return POLICY == GimbalPolicyType.ALLOWED
    }

    fun shouldClient(capability: Capability): Boolean {
        return EditorUtil.shouldPlayer(
            capability, EditorState(
                CURRENT_MODE,
                CURRENT_MODE_MODIFIER
            )
        )
    }

    fun isInNonDefaultMode(): Boolean {
        return CURRENT_MODE != Mode.NORMAL
    }

    /*
    * Send an updated player state to the server
    * */
    private fun updateServerState() {
        val state = this.getClientState()
        GimbalClient.sendStatePacket(state)
        ServerEditorManager.updateEditorState(MinecraftClient.getInstance().player!!.uuid, state)
    }


    /*
    * Set the current mode
    * */
    fun setMode(mode: Mode) {
        if (!isAllowed()) {
            ChatUtil.showToastToSelf("Gimble is disabled", "You cannot change modes", MinecraftClient.getInstance())
            return
        }
        if (!MinecraftClient.getInstance().player!!.isCreative)
            return
        CURRENT_MODE = mode

        onUpdated()
    }

    /*
    * Toggle a mode modifier
    * */
    fun toggleModifier(mod: ModeModifier) {
        if (!isAllowed()) {
            ChatUtil.showToastToSelf("Gimble is disabled", "You cannot change modifiers", MinecraftClient.getInstance())
            return
        }
        if (!MinecraftClient.getInstance().player!!.isCreative)
            return

        if (CURRENT_MODE.incompatibleModifiers.contains(mod)){
            if (TEMP_DISABLED_MODIFIERS.contains(mod))
                TEMP_DISABLED_MODIFIERS.remove(mod)
            else
                TEMP_DISABLED_MODIFIERS.add(mod)
        }
        else {
            if (CURRENT_MODE_MODIFIER.contains(mod))
                CURRENT_MODE_MODIFIER.remove(mod)
            else
                CURRENT_MODE_MODIFIER.add(mod)
        }

        onUpdated()
    }

    private fun onUpdated(shouldUpdateServer: Boolean = true) {
        CURRENT_MODE_MODIFIER.sortBy {
            it.displayName
        }
        checkForIncompatibleModeModifiers(shouldUpdateServer)
    }

    /**
     * This runs on Mode updated
     */
    private fun checkForIncompatibleModeModifiers(shouldUpdateServer: Boolean = true) {
        if (TEMP_DISABLED_MODIFIERS.size > 0) {
            CURRENT_MODE_MODIFIER.addAll(
                TEMP_DISABLED_MODIFIERS
            )
            TEMP_DISABLED_MODIFIERS.clear()
        }

        val incompatibleModeModifiers = CURRENT_MODE.incompatibleModifiers
        for (mod in CURRENT_MODE_MODIFIER) {
            if (incompatibleModeModifiers.contains(mod)) {
                TEMP_DISABLED_MODIFIERS.add(mod)
            }
        }
        for (mod in TEMP_DISABLED_MODIFIERS) {
            if (CURRENT_MODE_MODIFIER.contains(mod)) { // This check is probably useless.
                CURRENT_MODE_MODIFIER.remove(mod)
            }
        }

        CURRENT_MODE_MODIFIER.sortBy {
            it.displayName
        }

        // Update State
        if (shouldUpdateServer)
            updateServerState()

        if(TEMP_DISABLED_MODIFIERS.isNotEmpty()) {
            ChatUtil.showToastToSelf(
                "${CURRENT_MODE.displayName} Mode disabled",
                getListAsString(TEMP_DISABLED_MODIFIERS), MinecraftClient.getInstance()
            )
        }
    }

    private fun getListAsString(list: List<ModeModifier>): String{
        if(list.isEmpty())
            return "Empty list"
        val sb = StringBuilder()
        for (mod in list) {
            sb.append(mod.displayName)
            sb.append(", ")
        }
        return sb.toString().dropLast(2)
    }

    /*
    * Get the display text to display in the HUD
    * */
    fun getModeDisplayText(): Text {
        val displayText = StringBuilder(CURRENT_MODE.displayName)
        if (CURRENT_MODE_MODIFIER.isNotEmpty() && isAllowed() && MinecraftClient.getInstance().player?.isCreative == true) {
            displayText.append(" [")
            for (i in CURRENT_MODE_MODIFIER.indices) {
                displayText.append(CURRENT_MODE_MODIFIER[i].displayName)
                if (i != CURRENT_MODE_MODIFIER.size - 1) {
                    displayText.append(", ")
                }
            }
            displayText.append("]")
        } else if (!isAllowed() || !MinecraftClient.getInstance().player?.isCreative!!) {
            displayText.clear()
            displayText.append("[GIMBLE DISABLED]")
        }
        return Text.of(displayText.toString())
    }

    fun getCurrentColor(): Int {
        return CURRENT_MODE.color
    }

    fun getClientState(): EditorState {
        return EditorState(CURRENT_MODE, CURRENT_MODE_MODIFIER)
    }
}
